/*
 * Copyright 2007-2008 Ben Boeckel <MathStuf@gmail.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SIGMOD_MAP
#define SIGMOD_MAP

// Sigmod includes
#include "Matrix.h"
#include "Object.h"

// Qt includes
#include <QtCore/QList>

namespace Sigmod
{
// Forward declarations
class MapEffect;
class MapTrainer;
class MapWarp;
class MapWildList;
class Sigmod;

class SIGMOD_EXPORT Map : public Object
{
    Q_OBJECT
    Q_ENUMS(Type)
    
    public:
        enum Type
        {
            Outdoor = 0,
            Dungeon = 1,
            Building = 2
        };
        static const QStringList TypeStr;
        
        Map(const Map& map);
        Map(const Sigmod* parent, const int id);
        Map(const Map& map, const Sigmod* parent, const int id);
        Map(const QDomElement& xml, const Sigmod* parent, const int id = -1);
        ~Map();
        
        void validate();
        
        void load(const QDomElement& xml);
        QDomElement save() const;
        
        void setName(const QString& name);
        void setFlyWarp(const int flyWarp);
        void setType(const Type type);
        
        QString name() const;
        int flyWarp() const;
        Type type() const;
        
        void setTile(const int row, const int column, const int tile);
        void insertColumn(const int column);
        void insertRow(const int row);
        void addColumn();
        void addRow();
        void deleteColumn(const int column);
        void deleteRow(const int row);
        
        const Matrix<int>* map() const;
        Matrix<int>* map();
        int tile(const int row, const int column) const;
        int width() const;
        int height() const;
        QPoint size() const;
        
        const MapEffect* effect(const int index) const;
        MapEffect* effect(const int index);
        const MapEffect* effectById(const int id) const;
        MapEffect* effectById(const int id);
        int effectIndex(const int id) const;
        int effectCount() const;
        MapEffect* newEffect();
        MapEffect* newEffect(const QDomElement& xml);
        MapEffect* newEffect(const MapEffect& effect);
        void deleteEffect(const int index);
        void deleteEffectById(const int id);
        
        const MapTrainer* trainer(const int index) const;
        MapTrainer* trainer(const int index);
        const MapTrainer* trainerById(const int id) const;
        MapTrainer* trainerById(const int id);
        int trainerIndex(const int id) const;
        int trainerCount() const;
        MapTrainer* newTrainer();
        MapTrainer* newTrainer(const QDomElement& xml);
        MapTrainer* newTrainer(const MapTrainer& trainer);
        void deleteTrainer(const int index);
        void deleteTrainerById(const int id);
        
        const MapWarp* warp(const int index) const;
        MapWarp* warp(const int index);
        const MapWarp* warpById(const int id) const;
        MapWarp* warpById(const int id);
        int warpIndex(const int id) const;
        int warpCount() const;
        MapWarp* newWarp();
        MapWarp* newWarp(const QDomElement& xml);
        MapWarp* newWarp(const MapWarp& warp);
        void deleteWarp(const int index);
        void deleteWarpById(const int id);
        
        const MapWildList* wildList(const int index) const;
        MapWildList* wildList(const int index);
        const MapWildList* wildListById(const int id) const;
        MapWildList* wildListById(const int id);
        int wildListIndex(const int id) const;
        int wildListCount() const;
        MapWildList* newWildList();
        MapWildList* newWildList(const QDomElement& xml);
        MapWildList* newWildList(const MapWildList& wildList);
        void deleteWildList(const int index);
        void deleteWildListById(const int id);
        
        Map& operator=(const Map& rhs);
    private:
        int newEffectId() const;
        MapEffect* newEffect(MapEffect* effect);
        
        int newTrainerId() const;
        MapTrainer* newTrainer(MapTrainer* trainer);
        
        int newWarpId() const;
        MapWarp* newWarp(MapWarp* warp);
        
        int newWildListId() const;
        MapWildList* newWildList(MapWildList* wildList);
        
        void clear();
        
        QString m_name;
        int m_flyWarp;
        Type m_type;
        Matrix<int> m_map;
        QList<MapEffect*> m_effects;
        QList<MapTrainer*> m_trainers;
        QList<MapWarp*> m_warps;
        QList<MapWildList*> m_wildLists;
};
}
Q_DECLARE_METATYPE(Sigmod::Map::Type)

#endif
