/*
 * Copyright 2007-2008 Ben Boeckel <MathStuf@gmail.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SIGMOD_MATRIX
#define SIGMOD_MATRIX

// Sigmod includes
#include "Global.h"

// Qt includes
#include <QtCore/QPoint>
#include <QtCore/QVector>

namespace Sigmod
{
template<class T> class SIGMOD_EXPORT Matrix
{
    public:
        Matrix();
        Matrix(const int height, const int width, const T& value = T());
        Matrix(const Matrix<T>& rhs);
        
        void addRow(const T& value = T());
        void addColumn(const T& value = T());
        void insertRow(const int row, const T& value = T());
        void insertColumn(const int column, const T& value = T());
        void deleteRow(const int row);
        void deleteColumn(const int column);
        
        void clear();
        
        void set(const int row, const int column, const T& value);
        void resize(const int newHeight, const int newWidth, const T& value = T());
        
        T at(const int row, const int column) const;
        QVector<T> row(const int row) const;
        QVector<T> column(const int column) const;
        
        int height() const;
        int width() const;
        QPoint size() const;
        
        T& operator()(const int row, const int column);
        T operator()(const int row, const int column) const;
        Matrix<T>& operator=(const Matrix<T>& rhs);
        bool operator==(const Matrix<T>& rhs) const;
        bool operator!=(const Matrix<T>& rhs) const;
    private:
        QVector< QVector<T> > m_matrix;
};

template<class T> Matrix<T>::Matrix()
{
}

template<class T> Matrix<T>::Matrix(const int height, const int width, const T& value)
{
    resize(height, width, value);
}

template<class T> Matrix<T>::Matrix(const Matrix<T>& rhs)
{
    *this = rhs;
}

template<class T> void Matrix<T>::addRow(const T& value)
{
    if (size() == QPoint(0, 0))
        m_matrix.append(QVector<T>(1, value));
    else
        m_matrix.append(QVector<T>(width(), value));
}

template<class T> void Matrix<T>::addColumn(const T& value)
{
    if (size() == QPoint(0, 0))
        m_matrix.append(QVector<T>(1, value));
    else
    {
        for (int i = 0; i < height(); ++i)
            m_matrix[i].append(value);
    }
}

template<class T> void Matrix<T>::insertRow(const int row, const T& value)
{
    if (size() == QPoint(0, 0))
        m_matrix.append(QVector<T>(1, value));
    else
    {
        Q_ASSERT(row <= height());
        m_matrix.insert(row, QVector<T>(width(), value));
    }
}

template<class T> void Matrix<T>::insertColumn(const int column, const T& value)
{
    if (size() == QPoint(0, 0))
        m_matrix.append(QVector<T>(1, value));
    else
    {
        Q_ASSERT(column <= width());
        for (int i = 0; i < height(); ++i)
            m_matrix[i].insert(column, value);
    }
}

template<class T> void Matrix<T>::deleteRow(const int row)
{
    if (height() == 1)
        m_matrix.clear();
    else
    {
        Q_ASSERT(row < height());
        m_matrix.remove(row);
    }
}

template<class T> void Matrix<T>::deleteColumn(const int column)
{
    if (width() == 1)
        m_matrix.clear();
    else
    {
        Q_ASSERT(column < width());
        for (int i = 0; i < height(); ++i)
            m_matrix[i].remove(column);
    }
}

template<class T> void Matrix<T>::clear()
{
    m_matrix.clear();
}

template<class T> void Matrix<T>::set(const int row, const int column, const T& value)
{
    Q_ASSERT(row < height());
    Q_ASSERT(column < width());
    (m_matrix[row])[column] = value;
}

template<class T> void Matrix<T>::resize(const int newHeight, const int newWidth, const T& value)
{
    while (height() < newHeight)
        addRow(value);
    while (newHeight < height())
        deleteRow(newHeight);
    while (width() < newWidth)
        addColumn(value);
    while (newWidth < width())
        deleteColumn(newWidth);
}

template<class T> T Matrix<T>::at(const int row, const int column) const
{
    Q_ASSERT(row < height());
    Q_ASSERT(column < width());
    return m_matrix.at(row).at(column);
}

template<class T> QVector<T> Matrix<T>::row(const int row) const
{
    Q_ASSERT(row < height());
    return m_matrix.at(row);
}

template<class T> QVector<T> Matrix<T>::column(const int column) const
{
    Q_ASSERT(column < width());
    QVector<T> col;
    foreach (QVector<T> row, m_matrix)
        col.append(row.at(column));
    return col;
}

template<class T> int Matrix<T>::height() const
{
    return m_matrix.size();
}

template<class T> int Matrix<T>::width() const
{
    if (m_matrix.size())
        return m_matrix[0].size();
    return 0;
}

template<class T> QPoint Matrix<T>::size() const
{
    return QPoint(height(), width());
}

template<class T> T& Matrix<T>::operator()(const int row, const int column)
{
    Q_ASSERT(row < height());
    Q_ASSERT(column < width());
    return m_matrix[row][column];
}

template<class T> T Matrix<T>::operator()(const int row, const int column) const
{
    return at(row, column);
}

template<class T> Matrix<T>& Matrix<T>::operator=(const Matrix<T>& rhs)
{
    if (this == &rhs)
        return *this;
    resize(rhs.height(), rhs.width());
    for (int i = 0; i < height(); ++i)
    {
        for (int j = 0; j < width(); ++j)
            m_matrix[i][j] = rhs.m_matrix[i][j];
    }
    return *this;
}

template<class T> bool Matrix<T>::operator==(const Matrix<T>& rhs) const
{
    if (this == &rhs)
        return true;
    return m_matrix == rhs.m_matrix;
    return true;
}

template<class T> bool Matrix<T>::operator!=(const Matrix<T>& rhs) const
{
    return !(*this == rhs);
}

}

#endif
