/*
 * Copyright 2007-2008 Ben Boeckel <MathStuf@gmail.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SIGMOD_OBJECT
#define SIGMOD_OBJECT

// Sigmod includes
#include "Fraction.h"
#include "Global.h"
#include "Matrix.h"
#include "Script.h"

// Qt includes
#include <QtCore/QBuffer>
#include <QtCore/QObject>
#include <QtCore/QPoint>
#include <QtCore/QVariant>
#include <QtCore/QVarLengthArray>
#include <QtGui/QImage>
#include <QtXml/QDomElement>

// C includes
#include <climits>

namespace Sigmod
{
class SIGMOD_EXPORT Object : public QObject
{
    Q_OBJECT
    
    public:
        Object(const Object& object);
        Object(const Object* parent, const int id);
        virtual ~Object();
        
        const Object* parent() const;
        const Object* sigmod() const;
        
        int id() const;
        QString className() const;
        
        static QDomDocument xml(const Object* object);
        
        virtual void validate() = 0;
        
        virtual void load(const QDomElement& xml) = 0;
        virtual QDomElement save() const = 0;
    signals:
        void valMessage(const QString& message) const;
        void valWarning(const QString& message) const;
        void valError(const QString& message) const;
        void warning(const QString& message) const;
        void error(const QString& message) const;
        
        void changed() const;
    protected:
        void setId(const int id);
        
        static QString unused(const QString& variable);
        static QString bounds(const QString& variable);
        static QString subclass(const QString& subclass, const int id);
        static QString subclass(const QString& subclass, const QString& name);
        
        template<typename T> static void loadValue(const QDomElement& xml, T* value);
        template<typename T, int S> static void loadArray(const QDomElement& xml, QVarLengthArray<T, S>* value);
        template<typename T> static void loadList(const QDomElement& xml, QList<T>* value);
        template<typename T> static void loadMatrix(const QDomElement& xml, Matrix<T>* value);
        
        template<typename T> static QDomElement saveValue(const QString& name, const T& value);
        template<typename T, int S> static QDomElement saveArray(const QString& name, const QVarLengthArray<T, S>& value);
        template<typename T> static QDomElement saveList(const QString& name, const QList<T>& value);
        template<typename T> static QDomElement saveMatrix(const QString& name, const Matrix<T>& value);
        
        virtual void clear();
    private:
        int m_id;
        const Object* m_parent;
};

inline Object::~Object()
{
}

template<typename T> inline void Object::loadValue(const QDomElement& xml, T* value)
{
    *value = QVariant(xml.firstChild().toText().data()).value<T>();
}

template<> inline void Object::loadValue<bool>(const QDomElement& xml, bool* value)
{
    *value = (xml.firstChild().toText().data() == "true");
}

template<> inline void Object::loadValue<int>(const QDomElement& xml, int* value)
{
    *value = xml.firstChild().toText().data().toInt();
}

template<> inline void Object::loadValue<Fraction>(const QDomElement& xml, Fraction* value)
{
    value->set(xml.attribute("numerator", "1").toInt(), xml.attribute("denominator", "1").toInt());
}

template<> inline void Object::loadValue<QPoint>(const QDomElement& xml, QPoint* value)
{
    value->setX(xml.attribute("x", "0").toInt());
    value->setY(xml.attribute("y", "0").toInt());
}

template<> inline void Object::loadValue<QImage>(const QDomElement& xml, QImage* value)
{
    *value = QImage::fromData(QByteArray::fromBase64(xml.firstChild().toText().data().toUtf8()));
}

template<> inline void Object::loadValue<QByteArray>(const QDomElement& xml, QByteArray* value)
{
    *value = QByteArray::fromBase64(xml.firstChild().toText().data().toUtf8());
}

template<> inline void Object::loadValue<Script>(const QDomElement& xml, Script* value)
{
    value->setInterpreter(xml.attribute("interpreter", ""));
    value->setScript(xml.firstChild().toText().data());
}

template<typename T, int S> inline void Object::loadArray(const QDomElement& xml, QVarLengthArray<T, S>* value)
{
    QDomElement node = xml.firstChildElement("element");
    while (!node.isNull())
    {
        int index = node.attribute("index", "-1").toInt();
        if ((0 <= index) && (index < value->size()))
            loadValue(node, value->data() + index);
        node = node.nextSiblingElement("element");
    }
}

template<typename T> inline void Object::loadList(const QDomElement& xml, QList<T>* value)
{
    QDomElement node = xml.firstChildElement("element");
    T element;
    while (!node.isNull())
    {
        loadValue(node, &element);
        if (!value->contains(element))
            value->append(element);
        node = node.nextSiblingElement("element");
    }
}

template<typename T> inline void Object::loadMatrix(const QDomElement& xml, Matrix<T>* value)
{
    int height = xml.attribute("height", "1").toInt();
    int width = xml.attribute("width", "1").toInt();
    value->resize(width, height);
    QDomElement node = xml.firstChildElement("element");
    while (!node.isNull())
    {
        int row = node.attribute("row", "-1").toInt();
        int column = node.attribute("column", "-1").toInt();
        if ((0 <= row) && (row < height) && (0 <= column) && (column < width))
            loadValue(node, &value->operator()(row, column));
        node = node.nextSiblingElement("element");
    }
}

template<typename T> inline QDomElement Object::saveValue(const QString& name, const T& value)
{
    QDomElement element = QDomDocument().createElement(name);
    element.appendChild(QDomDocument().createTextNode(QVariant::fromValue<T>(value).toString()));
    return element;
}

template<> inline QDomElement Object::saveValue<bool>(const QString& name, const bool& value)
{
    QDomElement element = QDomDocument().createElement(name);
    element.appendChild(QDomDocument().createTextNode(value ? "true" : "false"));
    return element;
}

template<> inline QDomElement Object::saveValue<int>(const QString& name, const int& value)
{
    QDomElement element = QDomDocument().createElement(name);
    element.appendChild(QDomDocument().createTextNode(QString::number(value)));
    return element;
}

template<> inline QDomElement Object::saveValue<QString>(const QString& name, const QString& value)
{
    QDomElement element = QDomDocument().createElement(name);
    element.appendChild(QDomDocument().createTextNode(value));
    return element;
}

template<> inline QDomElement Object::saveValue<Fraction>(const QString& name, const Fraction& value)
{
    QDomElement element = QDomDocument().createElement(name);
    element.setAttribute("numerator", value.numerator());
    element.setAttribute("denominator", value.denominator());
    return element;
}

template<> inline QDomElement Object::saveValue<QPoint>(const QString& name, const QPoint& value)
{
    QDomElement element = QDomDocument().createElement(name);
    element.setAttribute("x", value.x());
    element.setAttribute("y", value.y());
    return element;
}

template<> inline QDomElement Object::saveValue<QImage>(const QString& name, const QImage& value)
{
    QDomElement element = QDomDocument().createElement(name);
    QByteArray bytes;
    QBuffer* buffer = new QBuffer(&bytes);
    buffer->open(QIODevice::WriteOnly);
    value.save(buffer, "PNG");
    element.appendChild(QDomDocument().createTextNode(bytes.toBase64()));
    delete buffer;
    return element;
}

template<> inline QDomElement Object::saveValue<QByteArray>(const QString& name, const QByteArray& value)
{
    QDomElement element = QDomDocument().createElement(name);
    element.appendChild(QDomDocument().createTextNode(value.toBase64()));
    return element;
}

template<> inline QDomElement Object::saveValue<Script>(const QString& name, const Script& value)
{
    QDomElement element = QDomDocument().createElement(name);
    element.setAttribute("interpreter", value.interpreter());
    element.appendChild(QDomDocument().createCDATASection(value.script()));
    return element;
}

template<typename T, int S> inline QDomElement Object::saveArray(const QString& name, const QVarLengthArray<T, S>& value)
{
    QDomElement element = QDomDocument().createElement(name);
    for (int i = 0; i < value.size(); ++i)
    {
        QDomElement valueElement = saveValue("element", value[i]);
        valueElement.setAttribute("index", QString::number(i));
        element.appendChild(valueElement);
    }
    return element;
}

template<typename T> inline QDomElement Object::saveList(const QString& name, const QList<T>& value)
{
    QDomElement element = QDomDocument().createElement(name);
    for (int i = 0; i < value.size(); ++i)
        element.appendChild(saveValue("element", value.at(i)));
    return element;
}

template<typename T> inline QDomElement Object::saveMatrix(const QString& name, const Matrix<T>& value)
{
    QDomElement element = QDomDocument().createElement(name);
    element.setAttribute("height", value.height());
    element.setAttribute("width", value.width());
    for (int i = 0; i < value.height(); ++i)
    {
        for (int j = 0; j < value.width(); ++j)
        {
            QDomElement valueElement = saveValue("element", value(i, j));
            valueElement.setAttribute("row", QString::number(i));
            valueElement.setAttribute("column", QString::number(j));
            element.appendChild(valueElement);
        }
    }
    return element;
}

}

#endif
