/*
 * Copyright 2007-2008 Ben Boeckel <MathStuf@gmail.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// Header include
#include "Species.h"

// Sigmod includes
#include "Macros.h"
#include "Rules.h"
#include "Sigmod.h"
#include "SpeciesAbility.h"
#include "SpeciesItem.h"
#include "SpeciesMove.h"

// Qt includes
#include <QtCore/QSet>

const QStringList Sigmod::Species::StyleStr = QStringList() << "Fluctuating" << "Fading" << "Slow" << "Normal" << "Fast" << "Erratic";

Sigmod::Species::Species(const Species& species) :
        Object(species.parent(), species.id()),
        m_baseStat(ST_End_GSC),
        m_effortValue(ST_End_GSC)
{
    *this = species;
}

Sigmod::Species::Species(const Sigmod* parent, const int id) :
        Object(parent, id),
        m_name(""),
        m_baseStat(ST_End_GSC),
        m_effortValue(ST_End_GSC),
        m_growth(Normal),
        m_catchValue(0),
        m_runChance(1, 1),
        m_fleeChance(1, 1),
        m_itemChance(1, 1),
        m_encyclopediaNumber(INT_MAX),
        m_weight(0),
        m_height(0),
        m_frontMaleSprite(-1),
        m_backMaleSprite(-1),
        m_frontFemaleSprite(-1),
        m_backFemaleSprite(-1),
        m_skin(-1),
        m_encyclopediaEntry(""),
        m_genderFactor(1, 1),
        m_eggSpecies(INT_MAX),
        m_eggSteps(0),
        m_evolution("", "")
{
    for (int i = 0; i < m_baseStat.size(); ++i)
        m_baseStat[i] = 0;
    for (int i = 0; i < m_effortValue.size(); ++i)
        m_effortValue[i] = 0;
}

Sigmod::Species::Species(const Species& species, const Sigmod* parent, const int id) :
        Object(parent, id),
        m_baseStat(ST_End_GSC),
        m_effortValue(ST_End_GSC)
{
    *this = species;
}

Sigmod::Species::Species(const QDomElement& xml, const Sigmod* parent, const int id) :
        Object(parent, id),
        m_baseStat(ST_End_GSC),
        m_effortValue(ST_End_GSC)
{
    LOAD_ID();
    load(xml);
}

Sigmod::Species::~Species()
{
    clear();
}

void Sigmod::Species::validate()
{
    TEST_BEGIN();
    if (m_name.isEmpty())
        emit(error("Name is empty"));
    TEST_ARRAY(setEffortValue, effortValue, ST_End_GSC);
    TEST(setGrowth, growth);
    TEST(setRunChance, runChance);
    TEST(setFleeChance, fleeChance);
    TEST(setItemChance, itemChance);
    TEST(setWeight, weight);
    TEST(setFrontMaleSprite, frontMaleSprite);
    TEST(setBackMaleSprite, backMaleSprite);
    TEST(setFrontFemaleSprite, frontFemaleSprite);
    TEST(setBackFemaleSprite, backFemaleSprite);
    TEST(setSkin, skin);
    if ((m_encyclopediaNumber != INT_MAX) && m_encyclopediaEntry.isEmpty())
        emit(error("Encyclopedia entry is empty"));
    TEST(setGenderFactor, genderFactor);
    if (qobject_cast<const Sigmod*>(sigmod())->rules()->breedingAllowed())
    {
        if (qobject_cast<const Sigmod*>(sigmod())->speciesIndex(m_eggSpecies) == INT_MAX)
            emit(error(bounds("egg species")));
        else
        {
            if (qobject_cast<const Sigmod*>(sigmod())->species(m_eggSpecies)->growth() != m_growth)
                emit(error("Growth mismatch with egg species"));
            TEST(setEggSteps, eggSteps);
        }
    }
    TEST_LIST(setType, type);
    TEST_LIST(setEggGroup, eggGroup);
    QSet<int> idChecker;
    QSet<int> valueChecker;
    if (abilityCount() < qobject_cast<const Sigmod*>(sigmod())->rules()->maxAbilities())
        emit(warning("There are too few abilities"));
    TEST_SUB_BEGIN(SpeciesAbility, abilities);
        TEST_SUB("ability", id);
        TEST_SUB_RAW(value, "ability", ability);
    TEST_SUB_END();
    idChecker.clear();
    valueChecker.clear();
    if (qobject_cast<const Sigmod*>(sigmod())->rules()->maxHeldItems() && !itemCount())
        emit(warning("There are no items"));
    TEST_SUB_BEGIN(SpeciesItem, items);
        TEST_SUB("item", id);
        TEST_SUB_RAW(value, "item", item);
    TEST_SUB_END();
    idChecker.clear();
    valueChecker.clear();
    if (!moveCount())
        emit(error("There are no moves"));
    TEST_SUB_BEGIN(SpeciesMove, moves);
        TEST_SUB("move", id);
        TEST_SUB_RAW(value, "move", move);
    TEST_SUB_END();
    TEST_END();
}

void Sigmod::Species::load(const QDomElement& xml)
{
    LOAD_BEGIN();
    LOAD(name);
    LOAD_ARRAY(baseStat);
    LOAD_ARRAY(effortValue);
    LOAD(growth);
    LOAD(experienceValue);
    LOAD(catchValue);
    LOAD(runChance);
    LOAD(fleeChance);
    LOAD(itemChance);
    LOAD(encyclopediaNumber);
    LOAD(weight);
    LOAD(height);
    LOAD(frontMaleSprite);
    LOAD(backMaleSprite);
    LOAD(frontFemaleSprite);
    LOAD(backFemaleSprite);
    LOAD(skin);
    LOAD(encyclopediaEntry);
    LOAD(genderFactor);
    LOAD(eggSpecies);
    LOAD(eggSteps);
    LOAD_LIST(type);
    LOAD_LIST(eggGroup);
    LOAD(evolution);
    LOAD_SUB(newAbility, SpeciesAbility);
    LOAD_SUB(newItem, SpeciesItem);
    LOAD_SUB(newMove, SpeciesMove);
}

QDomElement Sigmod::Species::save() const
{
    SAVE_CREATE();
    SAVE(name);
    SAVE_ARRAY(baseStat);
    SAVE_ARRAY(effortValue);
    SAVE(growth);
    SAVE(experienceValue);
    SAVE(catchValue);
    SAVE(runChance);
    SAVE(fleeChance);
    SAVE(itemChance);
    SAVE(encyclopediaNumber);
    SAVE(weight);
    SAVE(height);
    SAVE(frontMaleSprite);
    SAVE(backMaleSprite);
    SAVE(frontFemaleSprite);
    SAVE(backFemaleSprite);
    SAVE(skin);
    SAVE(encyclopediaEntry);
    SAVE(genderFactor);
    SAVE(eggSpecies);
    SAVE(eggSteps);
    SAVE_LIST( type);
    SAVE_LIST(eggGroup);
    SAVE(evolution);
    SAVE_SUB(SpeciesAbility, abilities);
    SAVE_SUB(SpeciesItem, items);
    SAVE_SUB(SpeciesMove, moves);
    return xml;
}

void Sigmod::Species::setName(const QString& name)
{
    CHECK(name);
}

void Sigmod::Species::setBaseStat(const Stat stat, const int baseStat)
{
    switch (stat)
    {
        case ST_SpecialDefense:
            if (qobject_cast<const Sigmod*>(sigmod())->rules()->specialSplit())
        case ST_HP ... ST_Special:
            break;
        default:
            emit(warning(bounds("stat")));
            return;
    }
    CHECK_ARRAY(baseStat[stat], baseStat);
}

void Sigmod::Species::setEffortValue(const Stat stat, const int effortValue)
{
    if (!qobject_cast<const Sigmod*>(sigmod())->rules()->effortValuesAllowed())
        emit(error("Effort Values not allowed"));
    else if (qobject_cast<const Sigmod*>(sigmod())->rules()->maxEVPerStat() < effortValue)
        emit(error(bounds("effortValue")));
    switch (stat)
    {
        case ST_SpecialDefense:
            if (qobject_cast<const Sigmod*>(sigmod())->rules()->specialSplit())
        case ST_HP ... ST_Special:
            break;
        default:
            emit(warning(bounds("stat")));
            return;
    }
    CHECK_ARRAY(effortValue[stat], effortValue);
}

void Sigmod::Species::setEffortValue(const int effortValue)
{
    if (qobject_cast<const Sigmod*>(sigmod())->rules()->effortValuesAllowed() && qobject_cast<const Sigmod*>(sigmod())->rules()->maxEVPerStat() < effortValue)
        emit(error(bounds("effortValue")));
}

void Sigmod::Species::setGrowth(const Style growth)
{
    CHECK(growth);
}

void Sigmod::Species::setExperienceValue(const int experienceValue)
{
    CHECK(experienceValue);
}

void Sigmod::Species::setCatchValue(const int catchValue)
{
    if (255 < catchValue)
        emit(error(bounds("catchValue")));
    else
        CHECK(catchValue);
}

void Sigmod::Species::setRunChance(const Fraction& runChance)
{
    if (1 < runChance)
        emit(error(bounds("runChance")));
    else
        CHECK(runChance);
}

void Sigmod::Species::setFleeChance(const Fraction& fleeChance)
{
    if (1 < fleeChance)
        emit(error(bounds("fleeChance")));
    else
        CHECK(fleeChance);
}

void Sigmod::Species::setItemChance(const Fraction& itemChance)
{
    if (1 < itemChance)
        emit(error(bounds("itemChance")));
    else
        CHECK(itemChance);
}

void Sigmod::Species::setEncyclopediaNumber(const int encyclopediaNumber)
{
    CHECK(encyclopediaNumber);
}

void Sigmod::Species::setWeight(const int weight)
{
    CHECK(weight);
}

void Sigmod::Species::setHeight(const int height)
{
    CHECK(height);
}

void Sigmod::Species::setEncyclopediaEntry(const QString& encyclopediaEntry)
{
    CHECK(encyclopediaEntry);
}

void Sigmod::Species::setFrontMaleSprite(const int frontMaleSprite)
{
    if (m_genderFactor.numerator() == m_genderFactor.denominator())
        emit(error("Cannot be male"));
    else if (qobject_cast<const Sigmod*>(sigmod())->spriteIndex(frontMaleSprite) == INT_MAX)
        emit(error(bounds("frontMaleSprite")));
    else
        CHECK(frontMaleSprite);
}

void Sigmod::Species::setBackMaleSprite(const int backMaleSprite)
{
    if (m_genderFactor.numerator() == m_genderFactor.denominator())
        emit(error("Cannot be male"));
    else if (qobject_cast<const Sigmod*>(sigmod())->spriteIndex(backMaleSprite) == INT_MAX)
        emit(error(bounds("backMaleSprite")));
    else
        CHECK(backMaleSprite);
}

void Sigmod::Species::setFrontFemaleSprite(const int frontFemaleSprite)
{
    if (!qobject_cast<const Sigmod*>(sigmod())->rules()->genderAllowed())
        emit(error("Gender is not allowed"));
    else if (!m_genderFactor.numerator())
        emit(error("Cannot be female"));
    else if (1 < m_genderFactor)
        emit(error("Genderless uses male sprites"));
    else if (qobject_cast<const Sigmod*>(sigmod())->spriteIndex(frontFemaleSprite) == INT_MAX)
        emit(error(bounds("frontFemaleSprite")));
    else
        CHECK(frontFemaleSprite);
}

void Sigmod::Species::setBackFemaleSprite(const int backFemaleSprite)
{
    if (!qobject_cast<const Sigmod*>(sigmod())->rules()->genderAllowed())
        emit(error("Gender is not allowed"));
    else if (!m_genderFactor.numerator())
        emit(error("Cannot be female"));
    else if (1 < m_genderFactor)
        emit(error("Genderless uses male sprites"));
    else if (qobject_cast<const Sigmod*>(sigmod())->spriteIndex(backFemaleSprite) == INT_MAX)
        emit(error(bounds("backFemaleSprite")));
    else
        CHECK(backFemaleSprite);
}

void Sigmod::Species::setSkin(const int skin)
{
    if (qobject_cast<const Sigmod*>(sigmod())->skinIndex(skin) == INT_MAX)
        emit(error(bounds("skin")));
    else
        CHECK(skin);
}

void Sigmod::Species::setGenderFactor(const Fraction& genderFactor)
{
    CHECK(genderFactor);
}

void Sigmod::Species::setEggSpecies(const int eggSpecies)
{
    if (qobject_cast<const Sigmod*>(sigmod())->speciesIndex(eggSpecies) == INT_MAX)
        emit(error(bounds("eggSpecies")));
    else
        CHECK(eggSpecies);
}

void Sigmod::Species::setEggSteps(const int eggSteps)
{
    CHECK(eggSteps);
}

void Sigmod::Species::setType(const int type, const bool state)
{
    if (qobject_cast<const Sigmod*>(sigmod())->typeIndex(type) == INT_MAX)
        emit(error(bounds("type")));
    else if (state && !m_type.contains(type))
    {
        m_type.append(type);
        emit(changed());
    }
    else if (m_type.contains(type))
    {
        m_type.removeAll(type);
        emit(changed());
    }
}

void Sigmod::Species::setEggGroup(const int eggGroup, const bool state)
{
    if (qobject_cast<const Sigmod*>(sigmod())->eggGroupIndex(eggGroup) == INT_MAX)
        emit(error(bounds("egg group")));
    else if (state && !m_eggGroup.contains(eggGroup))
    {
        m_eggGroup.append(eggGroup);
        emit(changed());
    }
    else if (m_eggGroup.contains(eggGroup))
    {
        m_eggGroup.removeAll(eggGroup);
        emit(changed());
    }
}

void Sigmod::Species::setEvolution(const Script& evolution)
{
    CHECK(evolution);
}

QString Sigmod::Species::name() const
{
    return m_name;
}

int Sigmod::Species::baseStat(const Stat stat) const
{
    switch (stat)
    {
        case ST_SpecialDefense:
            if (qobject_cast<const Sigmod*>(sigmod())->rules()->specialSplit())
        case ST_HP ... ST_Special:
            return m_effortValue[stat];
        default:
            emit(warning(bounds("base stat")));
            return -1;
    }
}

int Sigmod::Species::effortValue(const Stat stat) const
{
    switch (stat)
    {
        case ST_SpecialDefense:
            if (qobject_cast<const Sigmod*>(sigmod())->rules()->specialSplit())
        case ST_HP ... ST_Special:
            return m_effortValue[stat];
        default:
            emit(warning(bounds("effort value")));
            return -1;
    }
}

Sigmod::Species::Style Sigmod::Species::growth() const
{
    return m_growth;
}

int Sigmod::Species::experienceValue() const
{
    return m_experienceValue;
}

int Sigmod::Species::catchValue() const
{
    return m_catchValue;
}

Sigmod::Fraction Sigmod::Species::runChance() const
{
    return m_runChance;
}

Sigmod::Fraction Sigmod::Species::fleeChance() const
{
    return m_fleeChance;
}

Sigmod::Fraction Sigmod::Species::itemChance() const
{
    return m_itemChance;
}

int Sigmod::Species::encyclopediaNumber() const
{
    return m_encyclopediaNumber;
}

int Sigmod::Species::weight() const
{
    return m_weight;
}

int Sigmod::Species::height() const
{
    return m_height;
}

QString Sigmod::Species::encyclopediaEntry() const
{
    return m_encyclopediaEntry;
}

int Sigmod::Species::frontMaleSprite() const
{
    return m_frontMaleSprite;
}

int Sigmod::Species::backMaleSprite() const
{
    return m_backMaleSprite;
}

int Sigmod::Species::frontFemaleSprite() const
{
    return m_frontFemaleSprite;
}

int Sigmod::Species::backFemaleSprite() const
{
    return m_backFemaleSprite;
}

int Sigmod::Species::skin() const
{
    return m_skin;
}

Sigmod::Fraction Sigmod::Species::genderFactor() const
{
    return m_genderFactor;
}

int Sigmod::Species::eggSpecies() const
{
    return m_eggSpecies;
}

int Sigmod::Species::eggSteps() const
{
    return m_eggSteps;
}

bool Sigmod::Species::type(const int type) const
{
    return m_type.contains(type);
}

QList<int> Sigmod::Species::types() const
{
    return m_type;
}

bool Sigmod::Species::eggGroup(const int eggGroup) const
{
    return m_eggGroup.contains(eggGroup);
}

QList<int> Sigmod::Species::eggGroups() const
{
    return m_eggGroup;
}

Sigmod::Script Sigmod::Species::evolution() const
{
    return m_evolution;
}

const Sigmod::SpeciesAbility* Sigmod::Species::ability(const int index) const
{
    Q_ASSERT(index < abilityCount());
    return m_abilities.at(index);
}

Sigmod::SpeciesAbility* Sigmod::Species::ability(const int index)
{
    Q_ASSERT(index < abilityCount());
    return m_abilities[index];
}

const Sigmod::SpeciesAbility* Sigmod::Species::abilityById(const int id) const
{
    return ability(abilityIndex(id));
}

Sigmod::SpeciesAbility* Sigmod::Species::abilityById(const int id)
{
    return ability(abilityIndex(id));
}

int Sigmod::Species::abilityIndex(const int id) const
{
    for (int i = 0; i < abilityCount(); ++i)
    {
        if (m_abilities[i]->id() == id)
            return i;
    }
    return INT_MAX;
}

int Sigmod::Species::abilityCount() const
{
    return m_abilities.size();
}

Sigmod::SpeciesAbility* Sigmod::Species::newAbility()
{
    return newAbility(new SpeciesAbility(this, newAbilityId()));
}

Sigmod::SpeciesAbility* Sigmod::Species::newAbility(const QDomElement& xml)
{
    return newAbility(new SpeciesAbility(xml, this, newAbilityId()));
}

Sigmod::SpeciesAbility* Sigmod::Species::newAbility(const SpeciesAbility& ability)
{
    return newAbility(new SpeciesAbility(ability, this, newAbilityId()));
}

Sigmod::SpeciesAbility* Sigmod::Species::newAbility(SpeciesAbility* ability)
{
    m_abilities.append(ability);
    return ability;
}

void Sigmod::Species::deleteAbility(const int index)
{
    Q_ASSERT(index < abilityCount());
    delete m_abilities[index];
    m_abilities.removeAt(index);
}

void Sigmod::Species::deleteAbilityById(const int id)
{
    deleteAbility(abilityIndex(id));
}

int Sigmod::Species::newAbilityId() const
{
    int i = 0;
    while ((i < abilityCount()) && (abilityIndex(i) != INT_MAX))
        ++i;
    return i;
}

const Sigmod::SpeciesItem* Sigmod::Species::item(const int index) const
{
    Q_ASSERT(index < itemCount());
    return m_items.at(index);
}

Sigmod::SpeciesItem* Sigmod::Species::item(const int index)
{
    Q_ASSERT(index < itemCount());
    return m_items[index];
}

const Sigmod::SpeciesItem* Sigmod::Species::itemById(const int id) const
{
    return item(itemIndex(id));
}

Sigmod::SpeciesItem* Sigmod::Species::itemById(const int id)
{
    return item(itemIndex(id));
}

int Sigmod::Species::itemIndex(const int id) const
{
    for (int i = 0; i < itemCount(); ++i)
    {
        if (m_items[i]->id() == id)
            return i;
    }
    return INT_MAX;
}

int Sigmod::Species::itemCount() const
{
    return m_items.size();
}

Sigmod::SpeciesItem* Sigmod::Species::newItem()
{
    return newItem(new SpeciesItem(this, newItemId()));
}

Sigmod::SpeciesItem* Sigmod::Species::newItem(const QDomElement& xml)
{
    return newItem(new SpeciesItem(xml, this, newItemId()));
}

Sigmod::SpeciesItem* Sigmod::Species::newItem(const SpeciesItem& item)
{
    return newItem(new SpeciesItem(item, this, newItemId()));
}

Sigmod::SpeciesItem* Sigmod::Species::newItem(SpeciesItem* item)
{
    m_items.append(item);
    return item;
}

void Sigmod::Species::deleteItem(const int index)
{
    Q_ASSERT(index < itemCount());
    delete m_items[index];
    m_items.removeAt(index);
}

void Sigmod::Species::deleteItemById(const int id)
{
    deleteItem(itemIndex(id));
}

int Sigmod::Species::newItemId() const
{
    int i = 0;
    while ((i < itemCount()) && (itemIndex(i) != INT_MAX))
        ++i;
    return i;
}

const Sigmod::SpeciesMove* Sigmod::Species::move(const int index) const
{
    Q_ASSERT(index < moveCount());
    return m_moves.at(index);
}

Sigmod::SpeciesMove* Sigmod::Species::move(const int index)
{
    Q_ASSERT(index < moveCount());
    return m_moves[index];
}

const Sigmod::SpeciesMove* Sigmod::Species::moveById(const int id) const
{
    return move(moveIndex(id));
}

Sigmod::SpeciesMove* Sigmod::Species::moveById(const int id)
{
    return move(moveIndex(id));
}

int Sigmod::Species::moveIndex(const int id) const
{
    for (int i = 0; i < moveCount(); ++i)
    {
        if (m_moves[i]->id() == id)
            return i;
    }
    return INT_MAX;
}

int Sigmod::Species::moveCount() const
{
    return m_moves.size();
}

Sigmod::SpeciesMove* Sigmod::Species::newMove()
{
    return newMove(new SpeciesMove(this, newMoveId()));
}

Sigmod::SpeciesMove* Sigmod::Species::newMove(const QDomElement& xml)
{
    return newMove(new SpeciesMove(xml, this, newMoveId()));
}

Sigmod::SpeciesMove* Sigmod::Species::newMove(const SpeciesMove& move)
{
    return newMove(new SpeciesMove(move, this, newMoveId()));
}

Sigmod::SpeciesMove* Sigmod::Species::newMove(SpeciesMove* move)
{
    m_moves.append(move);
    return move;
}

void Sigmod::Species::deleteMove(const int index)
{
    Q_ASSERT(index < moveCount());
    delete m_moves[index];
    m_moves.removeAt(index);
}

void Sigmod::Species::deleteMoveById(const int id)
{
    deleteMove(moveIndex(id));
}

int Sigmod::Species::newMoveId() const
{
    int i = 0;
    while ((i < moveCount()) && (moveIndex(i) != INT_MAX))
        ++i;
    return i;
}

Sigmod::Species& Sigmod::Species::operator=(const Species& rhs)
{
    if (this == &rhs)
        return *this;
    clear();
    COPY(name);
    COPY(baseStat);
    COPY(effortValue);
    COPY(growth);
    COPY(experienceValue);
    COPY(catchValue);
    COPY(runChance);
    COPY(fleeChance);
    COPY(itemChance);
    COPY(encyclopediaNumber);
    COPY(weight);
    COPY(height);
    COPY(frontMaleSprite);
    COPY(backMaleSprite);
    COPY(frontFemaleSprite);
    COPY(backFemaleSprite);
    COPY(skin);
    COPY(encyclopediaEntry);
    COPY(genderFactor);
    COPY(eggSpecies);
    COPY(eggSteps);
    COPY(type);
    COPY(eggGroup);
    COPY(evolution);
    COPY_SUB(SpeciesAbility, abilities);
    COPY_SUB(SpeciesItem, items);
    COPY_SUB(SpeciesMove, moves);
    return *this;
}

void Sigmod::Species::clear()
{
    qDeleteAll(m_abilities);
    m_abilities.clear();
    qDeleteAll(m_items);
    m_items.clear();
    qDeleteAll(m_moves);
    m_moves.clear();
}
