/*
 * Copyright 2008 Ben Boeckel <MathStuf@gmail.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// Header include
#include "MapUI.h"

// Sigmodr includes
#include "TilemapModel.h"

// Sigmod includes
#include "../sigmod/Map.h"
#include "../sigmod/MapWarp.h"
#include "../sigmod/Sigmod.h"
#include "../sigmod/Sprite.h"
#include "../sigmod/Tile.h"

// Qt includes
#include <QtGui/QHeaderView>

Sigmodr::MapUI::MapUI(Sigmod::Map* map, QWidget* parent) :
        ObjectUI(parent)
{
    setupUi(this);
    setObjects(map, new Sigmod::Map(*map));
}

Sigmodr::MapUI::~MapUI()
{
}

void Sigmodr::MapUI::initGui()
{
    varType->addItem(Sigmod::Map::TypeStr[Sigmod::Map::Outdoor], QVariant::fromValue(Sigmod::Map::Outdoor));
    varType->addItem(Sigmod::Map::TypeStr[Sigmod::Map::Dungeon], QVariant::fromValue(Sigmod::Map::Dungeon));
    varType->addItem(Sigmod::Map::TypeStr[Sigmod::Map::Building], QVariant::fromValue(Sigmod::Map::Building));
    varTilemap->horizontalHeader()->setResizeMode(QHeaderView::Fixed);
    varTilemap->verticalHeader()->setResizeMode(QHeaderView::Fixed);
    varTilemap->horizontalHeader()->setDefaultSectionSize(64);
    varTilemap->verticalHeader()->setDefaultSectionSize(64);
    varTilemap->setModel(new TilemapModel(qobject_cast<Sigmod::Map*>(modified())->map(), sigmod()));
}

void Sigmodr::MapUI::refreshGui()
{
    varFlyWarp->clear();
    for (int i = 0; i < qobject_cast<Sigmod::Map*>(original())->warpCount(); ++i)
    {
        const Sigmod::MapWarp* warp = qobject_cast<Sigmod::Map*>(original())->warp(i);
        varFlyWarp->addItem(warp->name(), warp->id());
    }
    varTile->clear();
    for (int i = 0; i < sigmod()->tileCount(); ++i)
    {
        const Sigmod::Tile* tile = sigmod()->tile(i);
        if (sigmod()->spriteIndex(tile->sprite()) == INT_MAX)
            varTile->addItem(tile->name(), tile->id());
        else
            varTile->addItem(QPixmap::fromImage(sigmod()->spriteById(tile->sprite())->sprite()), tile->name(), tile->id());
    }
    varTile->setEnabled(false);
}

void Sigmodr::MapUI::setGui()
{
    varName->setText(qobject_cast<Sigmod::Map*>(modified())->name());
    boxFlyWarp->setChecked((qobject_cast<Sigmod::Map*>(modified())->flyWarp() == -1) ? Qt::Unchecked : Qt::Checked);
    varFlyWarp->setCurrentIndex(varFlyWarp->findData(qobject_cast<Sigmod::Map*>(modified())->flyWarp()));
    varType->setCurrentIndex(qobject_cast<Sigmod::Map*>(modified())->type());
    buttonDeleteColumn->setEnabled(0 < varTilemap->model()->columnCount());
    buttonDeleteRow->setEnabled(0 < varTilemap->model()->rowCount());
}

void Sigmodr::MapUI::apply()
{
    *qobject_cast<Sigmod::Map*>(original()) = *qobject_cast<Sigmod::Map*>(modified());
    emit(changed(false));
}

void Sigmodr::MapUI::discard()
{
    *qobject_cast<Sigmod::Map*>(modified()) = *qobject_cast<Sigmod::Map*>(original());
    setGui();
    emit(changed(false));
}

void Sigmodr::MapUI::on_varName_textChanged(const QString& name)
{
    const int cursor = varName->cursorPosition();
    qobject_cast<Sigmod::Map*>(modified())->setName(name);
    varName->setCursorPosition(cursor);
}

void Sigmodr::MapUI::on_boxFlyWarp_toggled()
{
    qobject_cast<Sigmod::Map*>(modified())->setFlyWarp(-1);
}

void Sigmodr::MapUI::on_varFlyWarp_activated(const int flyWarp)
{
    qobject_cast<Sigmod::Map*>(modified())->setFlyWarp(varFlyWarp->itemData(flyWarp).toInt());
}

void Sigmodr::MapUI::on_varType_activated(const int type)
{
    qobject_cast<Sigmod::Map*>(modified())->setType(varType->itemData(type).value<Sigmod::Map::Type>());
}

void Sigmodr::MapUI::on_varTilemap_clicked(const QModelIndex& index)
{
    m_index = index;
    varTile->setEnabled(true);
    varTile->setCurrentIndex(varTile->findData(varTilemap->model()->data(m_index, Qt::EditRole)));
}

void Sigmodr::MapUI::on_buttonAddColumn_pressed()
{
    varTilemap->model()->insertColumns(varTilemap->model()->columnCount(), 1);
    emit(changed(true));
}

void Sigmodr::MapUI::on_buttonAddRow_pressed()
{
    varTilemap->model()->insertRows(varTilemap->model()->rowCount(), 1);
    emit(changed(true));
}

void Sigmodr::MapUI::on_buttonDeleteColumn_pressed()
{
    varTilemap->model()->removeColumns(m_index.column(), 1);
    emit(changed(true));
}

void Sigmodr::MapUI::on_buttonDeleteRow_pressed()
{
    varTilemap->model()->removeRows(m_index.row(), 1);
    emit(changed(true));
}

void Sigmodr::MapUI::on_buttonInsertColumn_pressed()
{
    varTilemap->model()->insertColumns(m_index.column(), 1);
    emit(changed(true));
}

void Sigmodr::MapUI::on_buttonInsertRow_pressed()
{
    varTilemap->model()->insertRows(m_index.row(), 1);
    emit(changed(true));
}

void Sigmodr::MapUI::on_varTile_activated(const int tile)
{
    varTilemap->model()->setData(m_index, varTile->itemData(tile), Qt::EditRole);
    emit(changed(true));
    setGui();
}
